#ifndef WORLD_H
#define WORLD_H


#include <cmath>
#include <vector>
#include <helianthus.h>
#include <Box2D.h>


class World {
public:
	enum {
		CategoryGround = 1 << 0,
		CategoryObject = 1 << 1,
		CategoryMech   = 1 << 2,
	};
	
	enum {
		MaskGround     = 0xFFFF,
		MaskObject     = CategoryGround | CategoryObject,
		MaskMech       = CategoryGround,
	};
	
	class Child;
	typedef std::vector<Child*> ChildrenList;

private:
	static void eraseChild(ChildrenList &children, Child &child) {
		for(ChildrenList::reverse_iterator ri = children.rbegin(); ri != children.rend(); ++ri)
			if (*ri == &child) { children.erase((++ri).base()); return; }
	}

public:
	class Child {
	private:
		ChildrenList bases, children;
		void registerBase(Child &base)
			{ bases.push_back(&base); base.children.push_back(this); }
		void unregisterBase(Child &base)
			{ eraseChild(bases, base); eraseChild(base.children, *this); }
		
	public:
		World &world;
		explicit Child(World &world): world(world) { world.registerChild(*this); }
		explicit Child(Child &base): Child(base.world) { registerBase(base); }
		Child(const Child&) = delete;
		Child(Child &baseA, Child &baseB): Child(baseA) { registerBase(baseB); }
		Child(Child &baseA, Child &baseB, Child &baseC): Child(baseA, baseB) { registerBase(baseC); }
		virtual ~Child() {
			while(!children.empty()) delete children.back();
			while(!bases.empty()) unregisterBase(*bases.back());
			world.unregisterChild(*this);
		}
		
		virtual void draw() const {}
		virtual b2Body* pickBody(double /*x*/, double /*y*/) { return nullptr; }
	};
	
	const float minx, miny;
	const float maxx, maxy;
	const float gravity;
	const float stickWidth;
	const float motorWidth;
	
	b2World world;

private:
	float remainingTime;
	ChildrenList children;
	
	void registerChild(Child &child)
		{ children.push_back(&child); }
	void unregisterChild(Child &child)
		{ eraseChild(children, child); }
	
public:
	World(float minx, float miny, float maxx, float maxy, float gravity);
	~World();
	
	World(const World&) = delete;
	World& operator=(const World&) = delete;
	
	const ChildrenList& getChildren() const { return children; }
	
	void update(float dt);
	void draw() const;
};


#endif
